package fr.insa.clavardator.network;

import fr.insa.clavardator.util.ErrorHandler;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.net.*;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Objects;

public class NetDiscoverer {
	private static final short DISCOVERY_PORT = 31593;
	private static final short RESPONSE_PORT = 31594;
	private static final int BROADCAST_BUFFER_SIZE = 50;
	private static final int RESPONSE_BUFFER_SIZE = 50;

	private static DatagramSocket broadcastListener;
	private static DatagramSocket responseListener;
	private static boolean shouldStop = false;

	private NetDiscoverer() {
	}

	/**
	 * Discovers all active users in the network, and call the callback for each of them
	 *
	 * @param callback function to call when a new user is discovered
	 */
	public static void discoverActiveUsers(String broadcastMessage, ResponseReceivedCallback callback) {
		ResponseListener receiver = new ResponseListener(callback);
		BroadcastSender sender = new BroadcastSender(broadcastMessage);
		receiver.start();
		sender.start();
	}

	/**
	 * Starts to listen for discovery broadcasts and answers them
	 */
	public static void startDiscoveryListening(String responseMessage, @Nullable BroadcastReceivedCallback onBroadcastReceived) {
		BroadcastListener listener = new BroadcastListener((ipAddr, data) -> {
			if (onBroadcastReceived != null)
				onBroadcastReceived.onBroadcastReceived(ipAddr, data);
			new ResponseSender(ipAddr, responseMessage).start();
		}
		);
		listener.start();
	}

	public static void stopDiscovery() {
		shouldStop = true;
		if (broadcastListener != null)
			broadcastListener.close();
		if (responseListener != null)
			responseListener.close();
	}


	private static class BroadcastSender extends Thread {
		String broadcastMessage;

		/**
		 * Constructs and starts a thread that sends a broadcast over the network
		 *
		 * @param broadcastMessage The message to send
		 */
		public BroadcastSender(String broadcastMessage) {
			this.broadcastMessage = broadcastMessage;
		}

		@Override
		public void run() {
			byte[] buf = broadcastMessage.getBytes();
			try {
				for (InetAddress broadcastAddr : listAllBroadcastAddresses()) {
					DatagramSocket broadcastSocket = new DatagramSocket();
					broadcastSocket.setBroadcast(true);
					broadcastSocket.send(new DatagramPacket(buf, buf.length, broadcastAddr, DISCOVERY_PORT));
//					System.out.println("Broadcast sent with address " + broadcastAddr.toString());
				}

			} catch (IOException e) {
				ErrorHandler.getInstance().notifyError(e);
			}
		}
	}

	private static class BroadcastListener extends Thread {
		BroadcastReceivedCallback callback;

		public BroadcastListener(BroadcastReceivedCallback callback) {
			this.callback = callback;
		}

		@Override
		public void run() {
			while (!shouldStop) {
				try {
					broadcastListener = new DatagramSocket(null);
					broadcastListener.setOption(StandardSocketOptions.SO_REUSEPORT, true);
					broadcastListener.setOption(StandardSocketOptions.SO_REUSEADDR, true);
					broadcastListener.bind(new InetSocketAddress((InetAddress) null, DISCOVERY_PORT));

					byte[] buffer = new byte[BROADCAST_BUFFER_SIZE];
					DatagramPacket receivedPacket = new DatagramPacket(buffer, BROADCAST_BUFFER_SIZE);

					broadcastListener.receive(receivedPacket);
//					System.out.println("broadcast received from " + receivedPacket.getAddress().toString());
					callback.onBroadcastReceived(receivedPacket.getAddress(), new String(receivedPacket.getData()));
				} catch (IOException e) {
					if (!shouldStop) {
						ErrorHandler.getInstance().notifyError(e);
					}
				}
			}
		}
	}

	private static class ResponseSender extends Thread {
		String message;
		InetAddress address;

		/**
		 * Constructs and starts a thread that sends a broadcast over the network
		 *
		 * @param message The message to send
		 */
		public ResponseSender(InetAddress address, String message) {
			this.address = address;
			this.message = message;
		}

		@Override
		public void run() {
			byte[] buf = message.getBytes();
			try {
				DatagramSocket responseSocket = new DatagramSocket();
				responseSocket.send(new DatagramPacket(buf, buf.length, address, RESPONSE_PORT));
//				System.out.println("Response sent to " + address.toString());
			} catch (IOException e) {
				ErrorHandler.getInstance().notifyError(e);
			}
		}
	}

	private static class ResponseListener extends Thread {
		ResponseReceivedCallback callback;

		public ResponseListener(ResponseReceivedCallback callback) {
			this.callback = callback;
		}

		@Override
		public void run() {
			while (!shouldStop) {
				try {
					responseListener = new DatagramSocket(null);
					responseListener.setOption(StandardSocketOptions.SO_REUSEPORT, true);
					responseListener.setOption(StandardSocketOptions.SO_REUSEADDR, true);
					responseListener.bind(new InetSocketAddress((InetAddress) null, RESPONSE_PORT));

					byte[] buffer = new byte[RESPONSE_BUFFER_SIZE];
					DatagramPacket receivedPacket = new DatagramPacket(buffer, RESPONSE_BUFFER_SIZE);
					responseListener.receive(receivedPacket);
//					System.out.println("response received from " + receivedPacket.getAddress().toString());
					callback.onResponseReceived(receivedPacket.getAddress(), new String(receivedPacket.getData()));
				} catch (IOException e) {
					if (!shouldStop) {
						ErrorHandler.getInstance().notifyError(e);
					}
				}
			}
		}
	}


	static List<InetAddress> listAllBroadcastAddresses() throws SocketException {
		List<InetAddress> broadcastList = new ArrayList<>();
		Enumeration<NetworkInterface> interfaces = NetworkInterface.getNetworkInterfaces();

		while (interfaces.hasMoreElements()) {
			NetworkInterface networkInterface = interfaces.nextElement();
			if (!networkInterface.isLoopback() && networkInterface.isUp()) {
				networkInterface.getInterfaceAddresses().stream()
						.map(InterfaceAddress::getBroadcast)
						.filter(Objects::nonNull)
						.forEach(broadcastList::add);
			}
		}
		return broadcastList;
	}


	public interface ResponseReceivedCallback {
		void onResponseReceived(InetAddress ipAddr, String data);
	}

	private interface BroadcastReceivedCallback {
		void onBroadcastReceived(InetAddress ipAddr, String data);
	}

}
